import os
import json
import time
import sys
import warnings
import base64
from pathlib import Path
import re
import requests

redivis_dir = Path.home() / ".redivis"
cached_credentials = None
verify_ssl = (
    os.getenv("REDIVIS_API_ENDPOINT", "https://redivis.com").find(
        "https://localhost", 0
    )
    != 0
)
credentials_file = redivis_dir / "python_credentials"
default_scope = ["data.edit", "organization.write"]
client_id = "7YGtYWuQot1TEe0pHB3EPSj5"
# Note that https is optional, since traffic can happen over http if all in the same cluster
base_url = re.match(
    r"(https?://.*?)(/|$)", os.getenv("REDIVIS_API_ENDPOINT", "https://redivis.com")
).group(1)


def get_auth_token(scope=None):
    global cached_credentials

    if not scope:
        scope = default_scope

    if os.getenv("REDIVIS_API_TOKEN"):
        if os.getenv("REDIVIS_NOTEBOOK_JOB_ID") is None and bool(
            getattr(sys, "ps1", sys.flags.interactive)
        ):
            warnings.warn(
                """Setting the REDIVIS_API_TOKEN for interactive sessions is deprecated and highly discouraged.
Please delete the token on Redivis and remove it from your code, and follow the authentication prompts here instead.

This environment variable should only ever be set in a non-interactive environment, such as in an automated script or service.
"""
            )
        return os.environ["REDIVIS_API_TOKEN"]
    elif cached_credentials is None and credentials_file.is_file():
        try:
            with open(credentials_file, "r") as f:
                cached_credentials = json.load(f)
        except Exception as e:
            """ignore"""

    missing_scope = list(set(scope) - set(get_current_credential_scope()))

    if (
        cached_credentials is not None
        and "expires_at" in cached_credentials
        and "access_token" in cached_credentials
        and len(missing_scope) == 0
    ):
        if cached_credentials["expires_at"] < (time.time() - 5 * 60):
            return refresh_credentials()
        else:
            return cached_credentials["access_token"]
    else:
        if not redivis_dir.is_dir():
            redivis_dir.mkdir()

        perform_oauth_login(
            scope=missing_scope if len(missing_scope) > 0 else scope,
            upgrade_credentials=bool(len(missing_scope)),
        )

        return cached_credentials["access_token"]


def clear_cached_credentials():
    global cached_credentials
    cached_credentials = None
    credentials_file.unlink(missing_ok=True)


def perform_oauth_login(scope, amr_values=None, upgrade_credentials=False):
    global cached_credentials
    import webbrowser

    challenge, verifier = get_pkce()

    res = requests.post(
        f"{base_url}/oauth/device_authorization",
        verify=verify_ssl,
        headers={"Content-Type": "application/json"},
        data=json.dumps(
            {
                "client_id": client_id,
                "scope": " ".join(scope),
                "amr_values": amr_values,
                "code_challenge": challenge,
                "code_challenge_method": "S256",
                "access_type": "offline",
            }
        ),
    )
    res.raise_for_status()
    parsed_response = res.json()

    did_open = webbrowser.open(parsed_response["verification_uri_complete"])

    if did_open:
        print("Please authenticate with your Redivis account. Opening browser to:")
        print(parsed_response["verification_uri_complete"])
    else:
        print("Please visit the URL below to authenticate with your Redivis account:")
        print(parsed_response["verification_uri_complete"])

    headers = {}
    if upgrade_credentials and cached_credentials:
        headers["Authorization"] = f"Bearer {cached_credentials['access_token']}"

    started_polling_at = time.time()
    while True:
        if time.time() - started_polling_at > 60 * 10:
            raise Exception("Timed out waiting for device authorization")

        time.sleep(parsed_response["interval"] or 5)

        res = requests.post(
            f"{base_url}/oauth/token",
            verify=verify_ssl,
            headers=headers,
            data={
                "client_id": client_id,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "device_code": parsed_response["device_code"],
                "code_verifier": verifier,
            },
        )

        if res.status_code == 200:
            break
        elif res.status_code == 400:
            if res.json()["error"] == "authorization_pending":
                """authorization pending"""
            else:
                raise Exception(res.json())
        else:
            raise Exception(res.json())

    cached_credentials = res.json()

    with open(credentials_file, "w") as f:
        json.dump(cached_credentials, f, indent=2)

    return cached_credentials


def refresh_credentials(scope=None, amr_values=None):
    global cached_credentials

    if scope or amr_values:
        perform_oauth_login(
            scope=scope or get_current_credential_scope(),
            amr_values=amr_values,
            upgrade_credentials=True,
        )
    elif "refresh_token" in cached_credentials:
        res = requests.post(
            f"{base_url}/oauth/token",
            verify=verify_ssl,
            data={
                "client_id": client_id,
                "grant_type": "refresh_token",
                "refresh_token": cached_credentials["refresh_token"],
            },
        )
        if res.status_code >= 400:
            clear_cached_credentials()
        else:
            refresh_response = res.json()
            cached_credentials["access_token"] = refresh_response["access_token"]
            cached_credentials["expires_at"] = refresh_response["expires_at"]
            cached_credentials["expires_in"] = refresh_response["expires_in"]

            with open(credentials_file, "w") as f:
                json.dump(cached_credentials, f, indent=2)
    else:
        clear_cached_credentials()

    return get_auth_token()


def get_current_credential_scope():
    try:
        if cached_credentials is not None:
            base64_payload = cached_credentials["access_token"].split(".")[1]
            # IMPORTANT: b64decode requires that the string length be a multiple of 4, with "=" at the end for padding
            padded_base64_payload = f"{base64_payload}{'=' * (len(base64_payload) % 4)}"
            return json.loads(base64.b64decode(padded_base64_payload))["scope"].split(
                " "
            )
    except Exception as e:
        """ignore"""

    return default_scope


def get_pkce():
    import hashlib
    import base64
    import secrets

    verifier = secrets.token_urlsafe(64)
    challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(verifier.encode()).digest())
        .decode()
        .replace("=", "")
    )
    return challenge, verifier



import util
import pandas
import os


def test_linebreaks_in_cell():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    with open("tests/data/line_breaks.csv", "rb") as f:
        table.upload(name="test.csv").create(
            data=f,
            type="delimited",
            has_quoted_newlines=True,
        )


def test_upload_large_file():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )
    file_name = "concept_relationship.csv"
    with open(f"tests/data/{file_name}", "rb") as f:
        data = f.read()
        table.upload(name=file_name).create(data=data, wait_for_finish=True)


def test_upload_remove_on_failure():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    table.upload(
        name="tiny.csv",
    ).create(
        data='a,b\n1,2\n3,"4\n5"',
        remove_on_fail=True,
        type="delimited",
        has_quoted_newlines=False,
    )


def test_external_transfer():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    table.upload(
        name="test.csv",
    ).create(
        transfer_specification={
            "sourceType": "gcs",
            "sourcePath": "redivis-data-dev/test.csv.gz",
            "identity": "ian@redivis.com",
        },
    )


def test_upload_string():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    table.upload(
        name="tiny.csv",
    ).create(
        data='a,b\n1,2\n3,"4\n5"',
        type="delimited",
        has_quoted_newlines=True,
    )


def test_upload_large_string():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    with open("tests/data/us_counties_500k.geojson", "rb") as f:
        data = f.read()
        table.upload(name="us_counties_500k.geojson").create(
            data=data, wait_for_finish=True
        )


def test_upload_and_release():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(
        description="Some info", upload_merge_strategy="replace"
    )

    with open("tests/data/tiny.csv", "rb") as f:
        table.upload(name="tiny.csv").create(
            data=f, type="delimited", wait_for_finish=True
        )

    dataset.release()

    dataset = dataset.create_next_version()
    table = dataset.table(util.get_table_name())

    with open("tests/data/tiny.csv", "rb") as f:
        table.upload(name="tiny.csv").create(data=f, type="delimited")

    dataset.release()


def test_stream_bad_variable_name():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    table.create(description="Some info", upload_merge_strategy="replace")
    upload = table.upload(name="test")
    upload.create(type="stream", schema=[{"name": "var1", "type": "dateTime"}])

    upload2 = table.upload(name="test2")
    upload2.create(
        type="stream",
        schema=[
            {"name": "_upload_name", "type": "dateTime"},
            {"name": "var.2", "type": "dateTime"},
        ],
    )


def test_streaming_upload():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    table.create(description="Some info", upload_merge_strategy="replace")
    upload = table.upload(name="test")
    upload.create(type="stream")

    print(
        upload.insert_rows([{"var1": 1, "var2": "a", "var3": None}], update_schema=True)
    )

    upload.insert_rows([{"var2": 1}], update_schema=True)
    upload.insert_rows(
        [{"var1": "a", "var2": 1.2, "var3": "2020-01-01"}], update_schema=True
    )
    upload.insert_rows(
        [{"var1": "a", "var2": 1.2, "var3": "2020-01-01"}], update_schema=False
    )
    upload.insert_rows(
        [{"var1": "a", "var2": 1.2, "var3": "2020-01-01T00:00:00"}], update_schema=True
    )
    variables = upload.list_variables()
    print(variables)
    print(variables[0].get(wait_for_statistics=True))
    print(table.list_rows())
    print(upload.list_rows())


def test_redivis_upload():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    table.create(description="Some info", upload_merge_strategy="replace")
    upload = table.upload(name="test")
    upload.create(
        transfer_specification={
            "sourceType": "redivis",
            "sourcePath": "demo.novel_corona_virus_2019_dataset.covid_19_data",
        }
    )


def test_streaming_schema_upload():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    table.create(description="Some info", upload_merge_strategy="replace")
    upload = table.upload(name="test")
    upload.create(
        type="stream",
        schema=[
            {"name": "var1", "type": "integer"},
            {"name": "var2", "type": "integer"},
        ],
    )

    print(upload.insert_rows([{"var1": 1}], update_schema=False))

    upload.insert_rows([{"var2": 1}], update_schema=False)
    upload.insert_rows([{"var1": "a", "var2": 1.2}], update_schema=True)

    variables = upload.list_variables()
    print(variables)
    print(variables[0].get(wait_for_statistics=True))
    print(table.list_rows())
    print(upload.list_rows())


def test_resumable_upload():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()
    table.create(description="Some info", upload_merge_strategy="replace")

    with open(
        os.path.join(os.path.dirname(__file__), "./data/us_counties_500k.geojson"), "rb"
    ) as f:
        table.upload(name="us_counties_500k.geojson").create(type="geojson", data=f)

    print(table.upload("us_counties_500k.geojson").to_pandas_dataframe(10))


def test_streaming_performance():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    df = pandas.read_csv(os.path.join(os.path.dirname(__file__), "data/tiny.csv"))
    data = df.to_dict(orient="records")
    table.create(description="Some info")
    upload = table.upload(name="test")
    upload.create(type="stream")
    import time

    start_time = time.time()

    for x in range(20):
        upload.insert_rows(data, update_schema=True)

    print("--- %s seconds ---" % (time.time() - start_time))


def test_streaming_after_upload():
    util.create_test_dataset()
    util.clear_test_data()

    table = util.get_table()

    df = pandas.read_csv(os.path.join(os.path.dirname(__file__), "data/tiny.csv"))
    data = df.to_dict(orient="records")
    table.create(description="Some info")

    with open("tests/data/tiny.csv", "rb") as f:
        upload = table.upload(name="seed_file").create(data=f, type="delimited")

    upload = table.upload(name="streamed_data").create(
        type="stream",
        schema=[
            {
                "name": "id",
                "type": "integer",
            },
            {
                "name": "Numeric",
                "type": "float",
            },
            {
                "name": "Key",
                "type": "string",
            },
            {
                "name": "Timestamp",
                "type": "string",
            },
            {
                "name": "Year",
                "type": "integer",
            },
            {
                "name": "Boolean",
                "type": "boolean",
            },
            {
                "name": "Text",
                "type": "string",
            },
        ],
    )
    upload.insert_rows(data, update_schema=False)


def test_upload_raw_file():
    raise Exception("not implemented")
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create(is_file_index=True)

    with open("tests/data/tiny.csv", "rb") as f:
        file = table.add_file(name="tiny.csv", data=f)
        print(file)


def test_upload_metadata():
    dataset = util.create_test_dataset()
    util.clear_test_data()
    table = util.get_table().create()

    with open("tests/data/tiny.csv", "rb") as f:
        upload = table.upload("tiny.csv").create(
            f,
            metadata={
                "id": {
                    "label": "foo",
                    "description": "bar",
                    "valueLabels": {"1": "hello world"},
                }
            },
        )


def test_next_version_returned_from_create_next_version():
    # This makes sure we always get the next version of the dataset
    dataset = util.create_test_dataset()
    util.clear_test_data()
    dataset.release()
    next_version = dataset.create_next_version()
    assert "next" in next_version.uri
    dataset = util.get_dataset()
    next_version = dataset.create_next_version(if_not_exists=True)
    assert "next" in next_version.uri



from .Dataset import Dataset
from .Member import Member
from .Base import Base
from urllib.parse import quote as quote_uri
from ..common.api_request import make_paginated_request


class Organization(Base):
    def __init__(self, name):
        self.name = name
        self.uri = f"/organizations/{quote_uri(self.name, '')}"
        self.properties = {"kind": "organization", "uri": self.uri, "userName": name}

    def dataset(self, name, *, version=None):
        return Dataset(name, organization=self, version=version)

    def member(self, name):
        return Member(name, organization=self)

    def list_datasets(self, max_results=None, labels=None):
        query = {}
        if labels:
            query["labels"] = ",".join(labels)

        datasets = make_paginated_request(
            path=f"{self.uri}/datasets",
            page_size=100,
            max_results=max_results,
            query=query,
        )
        return [
            Dataset(dataset["name"], organization=self, properties=dataset)
            for dataset in datasets
        ]

    def list_members(self, max_results=None, labels=None, statuses=None):
        query = {}
        if labels:
            query["labels"] = ",".join(labels)
        if statuses:
            query["statuses"] = ",".join(statuses)

        members = make_paginated_request(
            path=f"{self.uri}/members",
            page_size=500,
            max_results=max_results,
            query=query,
        )
        return [
            Member(member["user"]["name"], organization=self, properties=member)
            for member in members
        ]



class Base:
    def __getitem__(self, key):
        return (
            self.properties[key]
            if hasattr(self, "properties")
            and self.properties is not None
            and key in self.properties
            else None
        )

    def __repr__(self) -> str:
        field_strings = []
        for key, field in vars(self).items():
            if (
                key != "properties"
                and key != "scoped_reference"
                and field is not None
                and not isinstance(field, Base)
            ):
                field_strings.append(f"{key}:{field!r}")

        return f"<{self.__class__.__name__} {' '.join(field_strings)}>"



from .Base import Base
import os
import time
import warnings

from ..common.api_request import make_request
from ..common.list_rows import list_rows
from ..common.util import get_geography_variable, get_warning, arrow_table_to_pandas


class Query(Base):
    def __init__(
        self,
        query,
        *,
        default_workflow=None,
        default_dataset=None,
    ):
        if not default_workflow and not default_dataset:
            if os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                default_workflow = os.getenv("REDIVIS_DEFAULT_WORKFLOW")
            elif os.getenv("REDIVIS_DEFAULT_DATASET"):
                default_dataset = os.getenv("REDIVIS_DEFAULT_DATASET")

        payload = {"query": query}
        if default_workflow:
            payload["defaultWorkflow"] = default_workflow
        if default_dataset:
            payload["defaultDataset"] = default_dataset

        self.properties = make_request(
            method="post",
            path="/queries",
            payload=payload,
        )
        self.uri = self.properties["uri"]

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def to_arrow_dataset(
        self,
        max_results=None,
        *,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_arrow_table(
        self,
        max_results=None,
        *,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_polars_lazyframe(
        self,
        max_results=None,
        *,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_dask_dataframe(
        self,
        max_results=None,
        *,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_pandas_dataframe(
        self,
        max_results=None,
        *,
        geography_variable="",
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

        return arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

    def to_geopandas_dataframe(
        self,
        max_results=None,
        *,
        geography_variable="",
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        import geopandas

        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

        df = arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                self.properties["outputSchema"], geography_variable
            )
            if geography_variable is None:
                raise Exception(
                    'Unable to find a variable with type=="geography" in the query results'
                )

        df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
            df[geography_variable["name"]]
        )
        df = geopandas.GeoDataFrame(
            data=df, geometry=geography_variable["name"], crs="EPSG:4326"
        )

        return df

    def to_dataframe(self, max_results=None, *, geography_variable="", progress=True):
        warnings.warn(get_warning("dataframe_deprecation"), FutureWarning, stacklevel=2)
        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                self.properties["outputSchema"], geography_variable
            )

        if geography_variable is not None:
            import geopandas

            warnings.warn(
                get_warning("geodataframe_deprecation"), FutureWarning, stacklevel=2
            )
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_arrow_batch_iterator(
        self, max_results=None, *, variables=None, progress=True
    ):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=False,
        )

    def list_rows(self, max_results=None, *, progress=True):
        warnings.warn(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.",
            FutureWarning,
            stacklevel=2,
        )

        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=self.properties["outputSchema"],
            output_type="tuple",
            progress=progress,
            coerce_schema=False,
        )

    def _wait_for_finish(self):
        while True:
            if self.properties["status"] == "completed":
                break
            elif self.properties["status"] == "failed":
                raise Exception(
                    f"Query job failed with message: {self.properties['errorMessage']}"
                )
            elif self.properties["status"] == "cancelled":
                raise Exception(f"Query job was cancelled")
            else:
                time.sleep(2)
                self.get()



import redivis

def test_organization_list_datasets():
    datasets = redivis.organization("Demo").list_datasets()
    print(datasets)
    print(datasets[0].properties)
    assert True

def test_user_list_datasets():
    datasets = redivis.user("imathews").list_datasets()
    print(datasets)
    assert True


import time
import os
import json
import logging
from urllib.parse import quote as quote_uri
import warnings
import io
from tqdm.auto import tqdm

from ..common.util import get_geography_variable, get_warning, arrow_table_to_pandas
from ..common.list_rows import list_rows

from .Base import Base
from .Variable import Variable
from ..common.api_request import make_request, make_paginated_request
from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload

MAX_SIMPLE_UPLOAD_SIZE = 1e7


class Upload(Base):
    def __init__(
        self,
        name,
        *,
        table,
        properties=None,
    ):
        self.table = table
        self.name = name
        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else (f"{table.uri}/uploads/{quote_uri(self.name,'')}")
        )
        self.properties = properties

    def create(
        self,
        data=None,
        *,
        type=None,
        transfer_specification=None,
        delimiter=None,
        schema=None,
        metadata=None,
        has_header_row=True,
        skip_bad_records=False,
        has_quoted_newlines=None,
        quote_character=None,
        escape_character=None,
        rename_on_conflict=False,
        replace_on_conflict=False,
        allow_jagged_rows=False,
        if_not_exists=False,
        remove_on_fail=False,
        wait_for_finish=True,
        raise_on_fail=True,
        progress=True,
    ):
        temp_upload_id = None

        if data and (
            (
                hasattr(data, "read")
                and os.stat(data.name).st_size > MAX_SIMPLE_UPLOAD_SIZE
            )
            or (hasattr(data, "__len__") and len(data) > MAX_SIMPLE_UPLOAD_SIZE)
        ):
            did_reopen_file = False
            pbar_bytes = None
            # If file isn't in binary mode, we need to reopen, otherwise uploading of non-text files will fail
            if hasattr(data, "mode") and "b" not in data.mode:
                data = open(data.name, "rb")
                did_reopen_file = True

            size = os.stat(data.name).st_size if hasattr(data, "read") else len(data)
            if progress:
                pbar_bytes = tqdm(total=size, unit="B", leave=False, unit_scale=True)

            res = make_request(
                method="POST",
                path=f"{self.table.uri}/tempUploads",
                payload={
                    "tempUploads": [
                        {"size": size, "name": self.name, "resumable": True}
                    ]
                },
            )
            temp_upload = res["results"][0]
            if temp_upload["resumable"]:
                perform_resumable_upload(
                    data=data,
                    progressbar=pbar_bytes,
                    temp_upload_url=temp_upload["url"],
                )
            else:
                perform_standard_upload(
                    data=data,
                    temp_upload_url=temp_upload["url"],
                    progressbar=pbar_bytes,
                )

            if did_reopen_file:
                data.close()
            if progress:
                pbar_bytes.close()

            temp_upload_id = temp_upload["id"]
            data = None
        elif data and not hasattr(data, "read"):
            data = io.StringIO(data)

        if schema and type != "stream":
            warnings.warn(
                "The schema option is ignored for uploads that aren't of type `stream`"
            )

        exists = self.exists()

        if if_not_exists and exists:
            return self

        if replace_on_conflict is True and rename_on_conflict is True:
            raise Exception(
                "Invalid parameters. replace_on_conflict and rename_on_conflict cannot both be True."
            )

        if exists:
            if replace_on_conflict is True:
                self.delete()
            elif rename_on_conflict is not True:
                raise Exception(
                    f"An upload with the name {self.name} already exists on this version of the table. If you want to upload this file anyway, set the parameter rename_on_conflict=True or replace_on_conflict=True."
                )

        files = None
        payload = {
            "name": self.name,
            "type": type,
            "schema": schema,
            "metadata": metadata,
            "hasHeaderRow": has_header_row,
            "skipBadRecords": skip_bad_records,
            "hasQuotedNewlines": has_quoted_newlines,
            "allowJaggedRows": allow_jagged_rows,
            "quoteCharacter": quote_character,
            "escapeCharacter": escape_character,
            "delimiter": delimiter,
            "tempUploadId": temp_upload_id,
            "transferSpecification": transfer_specification,
        }

        if data is not None:
            files = {"metadata": json.dumps(payload), "data": data}
            payload = None

        response = make_request(
            method="POST",
            path=f"{self.table.uri}/uploads",
            parse_payload=data is None,
            payload=payload,
            files=files,
        )
        self.properties = response
        self.uri = self.properties["uri"]

        try:
            if (data or temp_upload_id or transfer_specification) and wait_for_finish:
                while True:
                    time.sleep(2)
                    self.get()
                    if (
                        self.properties["status"] == "completed"
                        or self.properties["status"] == "failed"
                    ):
                        if self.properties["status"] == "failed" and raise_on_fail:
                            raise Exception(self.properties["errorMessage"])
                        break
                    else:
                        logging.debug("Upload is still in progress...")
        except Exception as e:
            if remove_on_fail:
                self.delete()
            raise Exception(str(e))

        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

    def insert_rows(self, rows, *, update_schema=False):
        response = make_request(
            method="POST",
            path=f"{self.uri}/rows",
            payload={"rows": rows, "updateSchema": update_schema},
        )
        return response

    def list_variables(self, *, max_results=10000):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], properties=variable, upload=self)
            for variable in variables
        ]

    def to_arrow_dataset(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_arrow_table(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_polars_lazyframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_dask_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

    def to_pandas_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )
        return arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

    def to_geopandas_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        geography_variable="",
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        import geopandas

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=True,
            batch_preprocessor=batch_preprocessor,
            max_parallelization=max_parallelization,
        )

        df = arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )
            if geography_variable is None:
                raise Exception(
                    'Unable to find a variable with type=="geography" in the query results'
                )

        if geography_variable is not None:
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_dataframe(
        self, max_results=None, *, variables=None, geography_variable="", progress=True
    ):
        warnings.warn(get_warning("dataframe_deprecation"), FutureWarning, stacklevel=2)

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=True,
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )

        if geography_variable is not None:
            import geopandas

            warnings.warn(
                get_warning("geodataframe_deprecation"), FutureWarning, stacklevel=2
            )
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_arrow_batch_iterator(
        self, max_results=None, *, variables=None, progress=True
    ):
        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=True,
        )

    def list_rows(self, max_results=None, *, variables=None, progress=True):
        warnings.warn(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.",
            FutureWarning,
            stacklevel=2,
        )

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="tuple",
            progress=progress,
            coerce_schema=True,
        )

    def variable(self, name):
        return Variable(name, upload=self)


def get_mapped_variables(variables, uri):
    all_variables = make_paginated_request(path=f"{uri}/variables", page_size=1000)

    if variables is None:
        return all_variables
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list



from .Table import Table
from .Base import Base
from ..common.api_request import make_request
from tqdm.auto import tqdm

import pathlib
import uuid
import os

from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload
from ..common.util import get_tempdir


class Notebook(Base):
    def __init__(
        self,
        current_notebook_job_id,
    ):
        self.current_notebook_job_id = current_notebook_job_id

    def create_output_table(
        self, data=None, *, name=None, append=False, geography_variables=None
    ):
        temp_file_path = f"{get_tempdir()}/out/{uuid.uuid4()}"
        try:
            pathlib.Path(temp_file_path).parent.mkdir(exist_ok=True, parents=True)

            import geopandas
            import pandas as pd
            import pyarrow as pa
            import pyarrow.dataset as pa_dataset
            import pyarrow.parquet as pa_parquet
            from dask.dataframe import DataFrame as dask_df

            if isinstance(data, geopandas.GeoDataFrame):
                if geography_variables is None:
                    geography_variables = list(data.select_dtypes("geometry"))
                data.to_wkt().to_parquet(path=temp_file_path, index=False)
            elif isinstance(data, pd.DataFrame):
                data.to_parquet(path=temp_file_path, index=False)
            elif isinstance(data, pa_dataset.Dataset):
                pa_dataset.write_dataset(
                    data,
                    temp_file_path,
                    format="parquet",
                    basename_template="part-{i}.parquet",
                    max_partitions=1,
                )
                temp_file_path = f"{temp_file_path}/part-0.parquet"
            elif isinstance(data, pa.Table):
                pa_parquet.write_table(data, temp_file_path)
            elif isinstance(data, dask_df):
                data.to_parquet(temp_file_path, write_index=False)
                temp_file_path = f"{temp_file_path}/part.0.parquet"
            else:
                # importing polars is causing an IllegalInstruction error on ARM + Docker. Import inline to avoid crashes elsewhwere
                # TODO: revert once fixed upstream
                import polars

                if isinstance(data, polars.LazyFrame):
                    data.sink_parquet(temp_file_path)
                elif isinstance(data, polars.DataFrame):
                    data.write_parquet(temp_file_path)
                else:
                    raise Exception(
                        "Unknown datatype provided to notebook.create_output_table. Must either by a file path, or an instance of pandas.DataFrame, pyarrow.Dataset, pyarrow.Table, dask.DataFrame, polars.LazyFrame, or polars.DataFrame"
                    )

            size = os.stat(temp_file_path).st_size

            pbar_bytes = tqdm(total=size, unit="B", leave=False, unit_scale=True)

            res = make_request(
                method="POST",
                path=f"/notebookJobs/{self.current_notebook_job_id}/tempUploads",
                payload={"tempUploads": [{"size": size, "resumable": size > 5e7}]},
            )
            temp_upload = res["results"][0]

            with open(temp_file_path, "rb") as f:
                if temp_upload["resumable"]:
                    perform_resumable_upload(
                        data=f,
                        progressbar=pbar_bytes,
                        proxy_url=f"{os.getenv('REDIVIS_API_ENDPOINT')}/notebookJobs/{self.current_notebook_job_id}/tempUploadProxy",
                        temp_upload_url=temp_upload["url"],
                    )
                else:
                    perform_standard_upload(
                        data=f,
                        temp_upload_url=temp_upload["url"],
                        proxy_url=f"{os.getenv('REDIVIS_API_ENDPOINT')}/notebookJobs/{self.current_notebook_job_id}/tempUploadProxy",
                        progressbar=pbar_bytes,
                    )

            pbar_bytes.close()
            os.remove(temp_file_path)

            res = make_request(
                method="PUT",
                path=f"/notebookJobs/{self.current_notebook_job_id}/outputTable",
                payload={
                    "name": name,
                    "append": append,
                    "geographyVariables": geography_variables,
                    "tempUploadId": temp_upload["id"],
                },
            )

            return Table(name=res["name"], properties=res)
        except Exception as e:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)
            raise e



from .Base import Base
from urllib.parse import quote as quote_uri
import warnings
import os
import time
import glob
import concurrent.futures
import pathlib
from tqdm.auto import tqdm
from threading import Event

from .Upload import Upload
from .Export import Export
from .Variable import Variable
from .File import File
from ..common.list_rows import list_rows
from ..common.api_request import make_request, make_paginated_request
from ..common.util import get_geography_variable, get_warning, arrow_table_to_pandas
from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload


class Table(Base):
    def __init__(
        self,
        name,
        *,
        dataset=None,
        workflow=None,
        properties=None,
    ):
        dataset, workflow = get_table_parents(dataset, workflow)
        self.name = name
        self.dataset = dataset
        self.workflow = workflow

        reference_scope = ""

        if dataset:
            reference_scope = f"{dataset.qualified_reference}."
        elif workflow:
            reference_scope = f"{workflow.qualified_reference}."

        self.qualified_reference = (
            properties["qualifiedReference"]
            if "qualifiedReference" in (properties or {})
            else (f"{reference_scope}{self.name}")
        )
        self.scoped_reference = (
            properties["scopedReference"]
            if "scopedReference" in (properties or {})
            else self.name
        )
        self.uri = f"/tables/{quote_uri(self.qualified_reference, '')}"
        self.properties = properties

    def create(
        self, *, description=None, upload_merge_strategy="append", is_file_index=False
    ):
        payload = {
            "name": self.name,
            "uploadMergeStrategy": upload_merge_strategy,
            "isFileIndex": is_file_index,
        }
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="POST",
            path=f"{self.dataset.uri}/tables",
            payload=payload,
        )
        update_properties(self, response)
        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def get(self):
        properties = make_request(method="GET", path=self.uri)
        update_properties(self, properties)
        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def list_uploads(self, *, max_results=None):
        uploads = make_paginated_request(
            path=f"{self.uri}/uploads", max_results=max_results
        )
        return [
            Upload(upload["name"], table=self, properties=upload) for upload in uploads
        ]

    def list_variables(self, *, max_results=None):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], table=self, properties=variable)
            for variable in variables
        ]

    def add_files(
        self,
        *,
        files=None,
        directory=None,
        progress=True,
        max_parallelization=os.cpu_count() * 5,
    ):
        if (files is None) == (directory is None):
            raise Exception("Either files or directory must be specified")

        total_size = 0

        if directory:
            files = []
            if not directory.endswith("/"):
                directory += "/"
            for filename in glob.iglob(directory + "**/*", recursive=True):
                if os.path.isfile(filename):
                    size = os.stat(filename).st_size
                    total_size += size
                    files.append(
                        {
                            "path": filename,
                            "name": os.path.relpath(filename, directory),
                            "size": size,
                        }
                    )
        else:
            files = list(map(map_file, files))
            for file in files:
                total_size += file["size"]

        pbar_bytes = None
        pbar_count = None
        if progress:
            pbar_count = tqdm(total=len(files), leave=False, unit=" files")
            pbar_bytes = tqdm(total=total_size, leave=False, unit="B", unit_scale=True)

        try:
            current_batch_timestamp = time.time()
            uploaded_files = []
            current_batch_files = []
            current_temp_uploads_batch = []
            target_batch_size = 1e8
            current_batch_size = 0

            for i, file in enumerate(files):
                if i % 1000 == 0:
                    res = make_request(
                        method="POST",
                        path=f"{self.uri}/tempUploads",
                        payload={
                            "tempUploads": [
                                {
                                    "size": file["size"],
                                    "name": file["name"],
                                    "resumable": file["size"] > 5e7,
                                }
                                for file in files[i : i + 1000]
                            ]
                        },
                    )
                    current_temp_uploads_batch = res["results"]

                current_batch_files.append(
                    {"file": file, "temp_upload": current_temp_uploads_batch[i % 1000]}
                )
                current_batch_size += file["size"]

                if (
                    len(current_batch_files) >= 1000
                    or i == len(files) - 1
                    or current_batch_size > target_batch_size
                ):

                    def upload(batch_file):
                        nonlocal pbar_bytes
                        file = batch_file["file"]
                        temp_upload = batch_file["temp_upload"]

                        if temp_upload["resumable"]:
                            data = (
                                open(file["path"], "rb")
                                if "path" in file
                                else file["data"]
                            )
                            perform_resumable_upload(
                                data=data,
                                progressbar=pbar_bytes,
                                temp_upload_url=temp_upload["url"],
                            )
                        else:
                            data = (
                                open(file["path"], "rb")
                                if "path" in file
                                else file["data"]
                            )
                            perform_standard_upload(
                                data=data,
                                temp_upload_url=temp_upload["url"],
                                progressbar=pbar_bytes,
                            )

                        if "path" in file:
                            data.close()

                    with concurrent.futures.ThreadPoolExecutor(
                        max_workers=min(100, max_parallelization)
                    ) as executor:
                        futures = [
                            executor.submit(upload, batch_file)
                            for batch_file in current_batch_files
                        ]

                        not_done = futures
                        try:
                            while not_done:
                                # next line 'sleeps' this main thread, letting the thread pool run
                                freshly_done, not_done = concurrent.futures.wait(
                                    not_done, timeout=0.5
                                )
                                for future in freshly_done:
                                    # Call result() on any finished threads to raise any exceptions encountered.
                                    future.result()
                        finally:
                            for future in not_done:
                                # Only cancels futures that were never started
                                future.cancel()
                            # Shutdown all background threads, now that they should know to exit early.
                            executor.shutdown(wait=True, cancel_futures=True)

                    if time.time() - current_batch_timestamp > 60:
                        target_batch_size = target_batch_size / 2
                    elif time.time() - current_batch_timestamp < 15:
                        target_batch_size = target_batch_size * 2

                    response = make_request(
                        method="POST",
                        path=f"{self.uri}/rawFiles",
                        payload={
                            "files": [
                                {
                                    "name": batch_file["file"]["name"],
                                    "tempUploadId": batch_file["temp_upload"]["id"],
                                }
                                for batch_file in current_batch_files
                            ]
                        },
                    )
                    uploaded_files += [
                        File(f["id"], table=self, properties=f)
                        for f in response["results"]
                    ]
                    if progress:
                        pbar_count.update(len(current_batch_files))

                    current_batch_timestamp = time.time()
                    current_batch_files = []
                    current_batch_size = 0

            if progress:
                pbar_bytes.close()
                pbar_count.close()

            return uploaded_files
        except Exception as e:
            if progress:
                pbar_bytes.close()
                pbar_count.close()

            raise e

    def download(
        self,
        path=None,
        *,
        format="csv",
        overwrite=False,
        progress=True,
    ):
        res = make_request(
            method="POST",
            path=f"{self.uri}/exports",
            payload={"format": format},
        )
        export_job = Export(res["id"], table=self, properties=res)
        return export_job.download_files(
            path=path,
            overwrite=overwrite,
            progress=progress,
        )

    def list_files(self, max_results=None, *, file_id_variable=None):
        if file_id_variable:
            variable = Variable(file_id_variable, table=self)
            if not variable.get().properties["isFileId"]:
                raise Exception(
                    f"The variable {file_id_variable} does not represent a file id."
                )
        else:
            variables = make_paginated_request(
                path=f"{self.uri}/variables", max_results=2, query={"isFileId": True}
            )
            if len(variables) == 0:
                raise Exception(
                    f"No variable containing file ids was found on this table"
                )
            elif len(variables) > 1:
                raise Exception(
                    f"This table contains multiple variables representing a file id. Please specify the variable with file ids you want to download via the 'file_id_variable' parameter."
                )

            file_id_variable = variables[0]["name"]

        rows = self.to_arrow_table(
            max_results=max_results, variables=[file_id_variable], progress=False
        ).to_pylist()
        return [File(row[file_id_variable], table=self) for row in rows]

    def download_files(
        self,
        path=None,
        *,
        overwrite=False,
        max_results=None,
        file_id_variable=None,
        progress=True,
        max_parallelization=os.cpu_count() * 5,
    ):
        files = self.list_files(max_results, file_id_variable=file_id_variable)
        if path is None:
            path = os.getcwd()

        if progress:
            pbar_count = tqdm(total=len(files), leave=False, unit=" files")
            pbar_bytes = tqdm(
                unit="B",
                leave=False,
                unit_scale=True,
                bar_format="{n_fmt} ({rate_fmt})",
            )

        if not os.path.exists(path):
            pathlib.Path(path).mkdir(exist_ok=True, parents=True)

        def on_progress(bytes):
            nonlocal pbar_bytes
            pbar_bytes.update(bytes)

        def download(file, cancel_event):
            nonlocal progress
            nonlocal pbar_count
            # TODO: if overwriting file, first check if file is the same size, and if so check md5 hash, and skip if identical (need md5 in header)
            file.download(
                path,
                overwrite=overwrite,
                progress=False,
                on_progress=on_progress if progress else None,
                cancel_event=cancel_event,
            )
            if progress and not cancel_event.is_set():
                pbar_count.update()

        cancel_event = Event()
        # TODO: this should use async, not threads
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=min(100, max_parallelization, len(files))
        ) as executor:
            futures = [executor.submit(download, file, cancel_event) for file in files]

            not_done = futures
            try:
                while not_done and not cancel_event.is_set():
                    # next line 'sleeps' this main thread, letting the thread pool run
                    freshly_done, not_done = concurrent.futures.wait(
                        not_done, timeout=0.5
                    )
                    for future in freshly_done:
                        # Call result() on any finished threads to raise any exceptions encountered.
                        future.result()
            except KeyboardInterrupt:
                print("KeyboardInterrupt")
            finally:
                cancel_event.set()

                for future in not_done:
                    # Only cancels futures that were never started
                    future.cancel()

                # Shutdown all background threads, now that they should know to exit early.
                executor.shutdown(wait=True, cancel_futures=True)

        if progress:
            pbar_count.close()
            pbar_bytes.close()

    def to_arrow_dataset(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )

    def to_arrow_table(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )

    def to_polars_lazyframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )

    def to_dask_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )

    def to_pandas_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )
        return arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

    def to_geopandas_dataframe(
        self,
        max_results=None,
        *,
        variables=None,
        geography_variable="",
        progress=True,
        dtype_backend="pyarrow",
        date_as_object=False,
        batch_preprocessor=None,
        max_parallelization=os.cpu_count(),
    ):
        import geopandas

        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self.properties["numBytes"]),
            max_parallelization=max_parallelization,
        )

        df = arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )
            if geography_variable is None:
                raise Exception(
                    'Unable to find a variable with type=="geography" in the query results'
                )

        if geography_variable is not None:
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_dataframe(
        self, max_results=None, *, variables=None, geography_variable="", progress=True
    ):
        warnings.warn(get_warning("dataframe_deprecation"), FutureWarning, stacklevel=2)

        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            use_export_api=should_use_export_api(self.properties["numBytes"]),
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )

        if geography_variable is not None:
            import geopandas

            warnings.warn(
                get_warning("geodataframe_deprecation"), FutureWarning, stacklevel=2
            )
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_arrow_batch_iterator(
        self, max_results=None, *, variables=None, progress=True
    ):
        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
        )

    def list_rows(self, max_results=None, *, variables=None, progress=True):
        warnings.warn(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.",
            FutureWarning,
            stacklevel=2,
        )

        if not self.properties or "container" not in self.properties:
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            table=self,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="tuple",
            progress=progress,
            coerce_schema=self.properties["container"]["kind"] == "dataset",
            use_export_api=should_use_export_api(self.properties["numBytes"]),
        )

    def update(self, *, name=None, description=None, upload_merge_strategy=None):
        payload = {}
        if name:
            payload["name"] = name
        if upload_merge_strategy:
            payload["uploadMergeStrategy"] = upload_merge_strategy
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        update_properties(self, response)
        return

    def upload(self, name):
        return Upload(name=name, table=self)

    def variable(self, name):
        return Variable(name, table=self)


def get_mapped_variables(variables, uri):
    all_variables = make_paginated_request(path=f"{uri}/variables", page_size=1000)
    if variables is None:
        return all_variables
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list


def get_table_parents(dataset, workflow):
    from .User import User

    if dataset or workflow:
        return dataset, workflow
    elif os.getenv("REDIVIS_NOTEBOOK_JOB_ID") is not None:
        return None, None
    elif os.getenv("REDIVIS_DEFAULT_WORKFLOW") is not None:
        return None, User(os.getenv("REDIVIS_DEFAULT_WORKFLOW").split(".")[0]).workflow(
            os.getenv("REDIVIS_DEFAULT_WORKFLOW").split(".")[1]
        )
    elif os.getenv("REDIVIS_DEFAULT_DATASET") is not None:
        return (
            User(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[0]).dataset(
                os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[1]
            )
        ), None
    else:
        raise Exception(
            "Cannot reference an unqualified table if the neither the REDIVIS_DEFAULT_WORKFLOW or REDIVIS_DEFAULT_DATASET environment variables are set."
        )


def update_properties(instance, properties):
    instance.properties = properties
    instance.qualified_reference = properties["qualifiedReference"]
    instance.scoped_reference = properties["scopedReference"]
    instance.name = properties["name"]
    instance.uri = properties["uri"]


def should_use_export_api(num_bytes):
    return num_bytes > (1e9 if os.getenv("REDIVIS_NOTEBOOK_JOB_ID") is None else 1e11)


def map_file(file):
    if isinstance(file, str):
        file = {"path": file}
    else:
        file = {**file}

    if "name" not in file:
        if "data" in file:
            raise Exception(
                'All file specifications with a "data" key must specify a name'
            )

        file["name"] = os.path.basename(file["path"])

    if "data" in file:
        if isinstance(file["data"], str):
            file["data"] = bytes(file["data"], "utf-8")
        file["size"] = len(file["data"])
    else:
        file["size"] = os.stat(file["path"]).st_size

    return file
